# ble_commands.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Builds byte level commands and defines parsing for anything sent over BLE."""

import enum
import logging
from abc import abstractmethod, ABC
from typing import Any, Callable, Dict, List

import wrapt
import betterproto
import construct  # For enum
from construct import Int8ub, Int32ub, Int64ub, Flag, GreedyString, GreedyBytes, Struct

from open_gopro import params
from open_gopro.constants import (
    ActionId,
    ErrorCode,
    QueryCmdId,
    UUID,
    SettingId,
    CmdId,
    StatusId,
    ProducerType,
)
from open_gopro.responses import GoProResp, Parser, ParserType, BuilderType, ParserBuilderType
from open_gopro import proto

logger = logging.getLogger(__name__)


class BLECommunicator(ABC):
    """Interface definition for a client to communicate via BLE.

    This interface is used to build commands parse responses for:
    - :py:class:`open_gopro.ble_commands.BleCommands`
    - :py:class:`open_gopro.ble_commands.BleSettings`
    - :py:class:`open_gopro.ble_commands.BleStatuses`
    """

    @abstractmethod
    def write(self, uuid: UUID, data: bytearray) -> GoProResp:
        """Write data to a UUID.

        Args:
            uuid (UUID): UUID to write to
            data (bytearray): data to write

        Returns:
            GoProResp: response
        """
        raise NotImplementedError

    @abstractmethod
    def read(self, uuid: UUID) -> GoProResp:
        """Read data from an UUID.

        Args:
            uuid (UUID): UUID to read from

        Returns:
            GoProResp: data read from UUID
        """
        raise NotImplementedError

    @abstractmethod
    def register_listener(self, producer: ProducerType) -> None:
        """Register to receive notifications for a producer.

        Args:
            producer (ProducerType): producer that we want to receive notifications from
        """
        raise NotImplementedError

    @abstractmethod
    def unregister_listener(self, producer: ProducerType) -> None:
        """Stop receiving notifications from a producer.

        Args:
            producer (ProducerType): Producer to stop receiving notifications for
        """
        raise NotImplementedError

    @abstractmethod
    def get_update(self, timeout: float = None) -> GoProResp:
        """Get a notification that was received from a registered producer.

        Args:
            timeout (float, optional): Time to wait for a notification before returning. Defaults to None (wait forever)

        Returns:
            GoProResp: the received update
        """
        raise NotImplementedError


# ========================================================Commands============================================


def read_cmd(uuid: UUID, response_parser: ParserType) -> Callable:
    """Build a BLE command that does a direct read from a UUID.

    Args:
        uuid (UUID): UUID to read from
        response_parser (ParserType): Parser to parse response. Can be either a construct Struct or a
        user defined Parser class

    Returns:
        Callable: function to build read command
    """

    @wrapt.decorator
    # pylint: disable = E, W
    def _wrapper(wrapped, instance, args, _):  # type: ignore
        logger.info(f"<----------- {wrapped.__name__} : {' '.join([str(a) for a in args])}")

        GoProResp.parser_map[uuid] = response_parser

        response = instance.communicator.read(uuid)
        logger.info(f"-----------> {wrapped.__name__} : {response}")
        return response

    return _wrapper


def write_cmd(
    id: CmdId,
    uuid: UUID,
    *,
    param_builder: BuilderType = None,
    response_parser: ParserType = None,
) -> Callable:
    """Build a BLE write command that writes to a UUID and receives responses via notifications.

    Args:
        id (CmdId): command to send
        uuid (UUID): uuid to write
        param_builder (BuilderType, optional): Optional builder that specifies how to build byte stream from param. Defaults to None.
        response_parser (ParserType, optional): Optional parser that defines how to build response from received byte stream. Defaults to None.

    Returns:
        Callable: function to build write command
    """

    @wrapt.decorator
    # pylint: disable = E, W
    def _wrapper(wrapped, instance, args, _):  # type: ignore
        status_struct = Struct("status" / construct.Enum(Int8ub, ErrorCode))

        logger.info(f"<----------- {wrapped.__name__} : {' '.join([str(a) for a in args])}")
        GoProResp.parser_map[id] = (
            status_struct if response_parser is None else (status_struct + response_parser)
        )

        data = bytearray([id.value])
        for arg in args:
            param = arg.value if issubclass(type(arg), enum.Enum) else arg
            # There must be a param builder if we have a param
            param = param_builder.build(param)  # type: ignore
            data.extend([len(param), *param])
        data.insert(0, len(data))
        response = instance.communicator.write(uuid, data)

        logger.info(f"-----------> \n{response}")
        return response

    return _wrapper


class ProtobufResponseAdapter(Parser):
    """Use a protobuf definition to parse a byte stream into a dict.

    Args:
        protobuf (Callable): protobuf to use as parser
    """

    def __init__(self, protobuf: betterproto.Message) -> None:
        self.protobuf = protobuf

    def parse(self, buf: bytearray) -> Dict[Any, Any]:
        """Parse byte stream into dict using protobuf.

        Args:
            buf (bytearray): input bytestream

        Returns:
            Dict[Any, Any]: output dict
        """
        return self.protobuf.FromString(buf).to_dict()


def proto_cmd(
    uuid: UUID, feature_id: CmdId, action_id: ActionId, request_proto: Callable, response_proto: Callable
) -> Callable:
    """Build a BLE command from a protobuf definition.

    Args:
        uuid (UUID): uuid to write to
        feature_id (CmdId): feature identifier
        action_id (ActionId): action identifier
        request_proto (Callable): protobuf definition to build request
        response_proto (Callable): protobuf definition to parse response

    Returns:
        Callable: function to build protobuf command
    """

    @wrapt.decorator
    # pylint: disable = E, W
    def _wrapper(wrapped, instance, args, kwargs):  # type: ignore
        logger.info(
            f"<----------- {wrapped.__name__} : {' '.join([*[str(a) for a in args], *[str(a) for a in kwargs.values()]])}"
        )

        if response_proto is not None:
            GoProResp.parser_map[action_id] = ProtobufResponseAdapter(response_proto)

        # Build request protobuf bytestream
        proto = request_proto()
        # Add args to protobuf request
        attrs = iter(wrapped.__annotations__.keys())
        for arg in args:
            param = arg.value if issubclass(type(arg), enum.Enum) else arg
            setattr(proto, next(attrs), param)
        # Add kwargs to protobuf request
        for name, arg in kwargs.items():
            if arg is not None:
                param = arg.value if issubclass(type(arg), enum.Enum) else arg
                setattr(proto, name, param)

        # Prepend headers and serialize
        request = bytearray([feature_id.value, action_id.value, *proto.SerializeToString()])
        # Prepend length
        request.insert(0, len(request))

        # Allow exception to pass through if protobuf not completely initialized
        response = instance.communicator.write(uuid, request)

        logger.info(f"-----------> \n{response}")
        return response

    return _wrapper


class BleCommands:
    """All of the BLE commands.

    To be used as a delegate for a BLECommunicator to build commands

    All of these return a GoProResp

    Args:
        communicator (BLECommunicator): [description]
    """

    def __init__(self, communicator: BLECommunicator):
        self.communicator = communicator

    @read_cmd(UUID.WAP_SSID, response_parser=Struct("ssid" / GreedyString("utf-8")))
    def get_wifi_ssid(self) -> GoProResp:
        """Get the Wifi SSID."""

    @read_cmd(UUID.WAP_PASSWORD, response_parser=Struct("password" / GreedyString("utf-8")))
    def get_wifi_password(self) -> GoProResp:
        """Get the WiFi password."""

    @write_cmd(CmdId.SET_SHUTTER, UUID.CQ_COMMAND, param_builder=Int8ub)
    def set_shutter(self, shutter: params.Shutter, /) -> GoProResp:
        """Set shutter on or off."""

    @write_cmd(CmdId.SET_PAIRING_COMPLETE, UUID.CQ_COMMAND, param_builder=Int8ub)
    def set_pairing_complete(self, status: bool, /) -> GoProResp:
        """Notify the peer that pairing has completed."""

    @write_cmd(CmdId.GET_CAMERA_STATUS, UUID.CQ_QUERY)
    def get_camera_status(self) -> GoProResp:
        """Get camera status."""

    @write_cmd(CmdId.GET_SETTINGS_JSON, UUID.CQ_COMMAND)
    def get_settings_json(self) -> GoProResp:
        """Get settings.json file."""

    @write_cmd(CmdId.GET_HW_INFO, UUID.CQ_COMMAND)
    def get_hardware_info(self) -> GoProResp:
        """Get hardware information."""

    @write_cmd(CmdId.LOAD_PRESET, UUID.CQ_COMMAND, param_builder=Int32ub)
    def load_preset(self, preset: params.Preset, /) -> GoProResp:
        """Load a Preset."""

    @write_cmd(CmdId.SET_WIFI, UUID.CQ_COMMAND, param_builder=Int8ub)
    def enable_wifi_ap(self, enable: bool, /) -> GoProResp:
        """Enable / disable the Wi-Fi Access Point."""

    @write_cmd(CmdId.SET_THIRD_PARTY_CLIENT_INFO, UUID.CQ_COMMAND)
    def set_third_party_client_info(self) -> GoProResp:
        """Flag as third party app."""

    @write_cmd(
        CmdId.GET_THIRD_PARTY_API_VERSION,
        UUID.CQ_COMMAND,
        response_parser=Struct("major" / Int8ub, "minor" / Int8ub),
    )
    def get_third_party_api_version(self) -> GoProResp:
        """Get Third party API version that is being used."""

    @proto_cmd(
        UUID.CQ_COMMAND,
        CmdId.SET_TURBO_MODE,
        ActionId.SET_TURBO_MODE,
        proto.RequestSetTurboActive,
        proto.ResponseGeneric,
    )
    def set_turbo_mode(self, active: bool, /) -> GoProResp:
        """Enable / disable turbo mode."""

    @proto_cmd(
        UUID.CQ_COMMAND,
        CmdId.GET_PRESET_STATUS,
        ActionId.GET_PRESET_STATUS,
        proto.RequestGetPresetStatus,
        proto.NotifyPresetStatus,
    )
    def get_preset_status(
        self,
        register_preset_status: List[params.EnumRegisterPresetStatus] = None,
        unregister_preset_status: List[params.EnumRegisterPresetStatus] = None,
    ) -> GoProResp:
        """Get the preset status.

        TODO This isn't working

        Args:
            register_preset_status (List[params.EnumRegisterPresetStatus], optional): [description]. Defaults to None.
            unregister_preset_status (List[params.EnumRegisterPresetStatus], optional): [description]. Defaults to None.

        Returns:
            GoProResp: the response
        """


# ========================================================Settings============================================


@wrapt.decorator
# pylint: disable = E, W
def setter_cmd(wrapped, instance, args, _):  # type: ignore
    """Build a BLE command that sets a Setting value."""
    logger.info(f"<----------- {wrapped.__name__} : {' '.join([str(a) for a in args])}")

    data = bytearray([instance.id.value])
    try:
        param = int(args[0])  # Do this first in separate step to allow IndexError to propagate
        param = instance.builder.build(param)  # There must be a param builder if we have a param
        data.extend([len(param), *param])
    except IndexError:
        pass
    data.insert(0, len(data))
    response = instance.communicator.write(instance.setter_uuid, data)

    logger.info(f"-----------> \n{response}")
    return response


@wrapt.decorator
# pylint: disable = E, W
def log_query(wrapped, instance, args, kwargs):  # type: ignore
    """Log a query write."""
    logger.info(f"<----------- {wrapped.__name__} : {instance.id}")
    response = wrapped(*args, **kwargs)
    logger.info(f"-----------> {response}")
    return response


class Setting(ABC):
    """An individual camera setting.

    Args:
        communicator (BLECommunicator): Adapter to read write settings data
        id (SettingId): ID of setting
        parser_builder (ParserBuilderType): object to both parse and build setting
    """

    def __init__(
        self, communicator: BLECommunicator, id: SettingId, parser_builder: ParserBuilderType
    ) -> None:
        self.id = id
        self.communicator: BLECommunicator = communicator
        self.setter_uuid: UUID = UUID.CQ_SETTINGS
        self.reader_uuid: UUID = UUID.CQ_QUERY
        self.parser: ParserBuilderType = parser_builder
        self.builder: ParserBuilderType = parser_builder  # Just syntactic sugar
        # Add to param parsing map
        GoProResp.parser_map[self.id] = self.parser

    @abstractmethod
    def set(self) -> GoProResp:
        """Set the value of the setting.

        This shall be implemented (and documented) by the subclass.

        Returns:
            GoProResp: Status of set
        """
        raise NotImplementedError("To be implemented by subclass")

    @log_query
    def get_value(self) -> GoProResp:
        """Get the settings value.

        Returns:
            GoProResp: settings value
        """
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.GET_SETTING_VAL))

    @log_query
    def get_name(self) -> GoProResp:
        """Get the settings name.

        Raises:
            NotImplementedError: This isn't implemented on the camera

        Returns:
            GoProResp: settings name
        """
        # return self.communicator.write(
        #     self.reader_uuid, QueryCmdId.GET_SETTING_NAME, self._build_cmd(QueryCmdId.GET_SETTING_NAME)
        # )
        raise NotImplementedError("Not implemented on camera!")

    @log_query
    def get_capabilities_values(self) -> GoProResp:
        """Get currently supported settings capabilities values.

        Returns:
            GoProResp: settings capabilities values
        """
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.GET_CAPABILITIES_VAL))

    @log_query
    def get_capabilities_names(self) -> GoProResp:
        """Get currently supported settings capabilities names.

        Raises:
            NotImplementedError: This isn't implemented on the camera

        Returns:
            GoProResp: settings capabilities names
        """
        # return self.communicator.write(
        #     self.reader_uuid,
        #     QueryCmdId.GET_CAPABILITIES_NAME,
        #     self._build_cmd(QueryCmdId.GET_CAPABILITIES_NAME),
        # )
        raise NotImplementedError("Not implemented on camera!")

    @log_query
    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Current value of respective setting ID
        """
        self.communicator.register_listener((QueryCmdId.SETTING_VAL_PUSH, self.id))
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.REG_SETTING_VAL_UPDATE))

    @log_query
    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator.unregister_listener((QueryCmdId.SETTING_VAL_PUSH, self.id))
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.UNREG_SETTING_VAL_UPDATE))

    @log_query
    def register_capability_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's capabilities update.

        Returns:
            GoProResp: Current capabilities of respective setting ID
        """
        self.communicator.register_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self.id))
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.REG_CAPABILITIES_UPDATE))

    @log_query
    def unregister_capability_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's capabilities change.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator.unregister_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self.id))
        return self.communicator.write(self.reader_uuid, self._build_cmd(QueryCmdId.UNREG_CAPABILITIES_UPDATE))

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data to send a settings query over-the-air.

        Args:
            cmd (QueryCmdId): command to build

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self.id.value])
        ret.insert(0, len(ret))
        return ret


# Lots of boiler-plate here. This is because we need a way to keep the type-hinting on the "set" method. And
# also to make the docstring generated documentation look nice.
class BleSettings:
    """The collection of all Settings.

    To be used by a BLECommunicator delegate to build setting commands.

    Args:
        communicator (BLECommunicator): Adapter to read / write settings
    """

    def __init__(self, communicator: BLECommunicator):
        self.communicator = communicator

        class Resolution(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, resolution: params.Resolution) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.resolution = Resolution(
            self.communicator, SettingId.RESOLUTION, construct.Enum(Int8ub, params.Resolution)
        )
        """Resolution. Set with :py:class:`open_gopro.params.Resolution`"""

        class FOV(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, fov: params.FieldOfView) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.video_field_of_view = FOV(
            self.communicator, SettingId.VIDEO_FOV, construct.Enum(Int8ub, params.FieldOfView)
        )
        """Video FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        self.photo_field_of_view = FOV(
            self.communicator, SettingId.PHOTO_FOV, construct.Enum(Int8ub, params.FieldOfView)
        )
        """Photo FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        self.multi_shot_field_of_view = FOV(
            self.communicator, SettingId.MULTI_SHOT_FOV, construct.Enum(Int8ub, params.FieldOfView)
        )
        """Multi-shot FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        class FPS(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, fps: params.FPS) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.fps = FPS(self.communicator, SettingId.FPS, construct.Enum(Int8ub, params.FPS))
        """Frames per second. Set with :py:class:`open_gopro.params.FPS`"""

        class LED(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, led: params.LED) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.led = LED(self.communicator, SettingId.LED, construct.Enum(Int8ub, params.LED))
        """Set the LED options (or also send the BLE keep alive signal). Set with :py:class:`open_gopro.params.LED`"""

        class MaxLensMode(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, mode: params.MaxLensMode) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.max_lens_mode = MaxLensMode(
            self.communicator, SettingId.MAX_LENS_MOD, construct.Enum(Int8ub, params.MaxLensMode)
        )
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.params.MaxLensMode`"""

        class Shortcut(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, shortcut: params.Shortcut) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.lower_left_shortcut = Shortcut(
            self.communicator, SettingId.SHORTCUT_LOWER_LEFT, construct.Enum(Int8ub, params.Shortcut)
        )
        """Lower left shortcut. Set with :py:class:`open_gopro.params.Shortcut`"""

        self.lower_right_shortcut = Shortcut(
            self.communicator, SettingId.SHORTCUT_LOWER_RIGHT, construct.Enum(Int8ub, params.Shortcut)
        )
        """Lower right shortcut. Set with :py:class:`open_gopro.params.Shortcut`"""

        self.upper_left_shortcut = Shortcut(
            self.communicator, SettingId.SHORTCUT_UPPER_LEFT, construct.Enum(Int8ub, params.Shortcut)
        )
        """Upper left shortcut. Set with :py:class:`open_gopro.params.Shortcut`"""

        self.upper_right_shortcut = Shortcut(
            self.communicator, SettingId.SHORTCUT_UPPER_RIGHT, construct.Enum(Int8ub, params.Shortcut)
        )
        """Upper right shortcut. Set with :py:class:`open_gopro.params.Shortcut`"""


# =================================================Statuses============================================


class Status:
    """An individual camera status.

    Args:
        communicator (BLECommunicator): Adapter to read status data
        id (StatusId): ID of status
    """

    uuid = UUID.CQ_QUERY

    def __init__(self, communicator: BLECommunicator, id: StatusId, parser: ParserType) -> None:
        self.id = id
        self.communicator = communicator
        self.parser = parser
        # Add to response parsing map
        GoProResp.parser_map[self.id] = self.parser

    @log_query
    def get_value(self) -> GoProResp:
        """Get the current value of a status.

        Returns:
            GoProResp: current status value
        """
        return self.communicator.write(Status.uuid, self._build_cmd(QueryCmdId.GET_STATUS_VAL))

    @log_query
    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a status changes.

        Returns:
            GoProResp: current status value
        """
        self.communicator.register_listener((QueryCmdId.STATUS_VAL_PUSH, self.id))
        return self.communicator.write(Status.uuid, self._build_cmd(QueryCmdId.REG_STATUS_VAL_UPDATE))

    @log_query
    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when status changes.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator.unregister_listener((QueryCmdId.STATUS_VAL_PUSH, self.id))
        return self.communicator.write(Status.uuid, self._build_cmd(QueryCmdId.UNREG_STATUS_VAL_UPDATE))

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data for a given status command.

        Args:
            cmd (QueryCmdId): command to build data for

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self.id.value])
        ret.insert(0, len(ret))
        return ret


class BleStatuses:
    """All of the BLE Statuses.

    To be used by a BLECommunicator delegate to build status commands.
    """

    def __init__(self, communicator: BLECommunicator) -> None:
        self.communicator = communicator

        self.batt_present: Status = Status(self.communicator, StatusId.BATT_PRESENT, Flag)  # type: ignore
        """Is the system's internal battery present?"""

        self.batt_level: Status = Status(self.communicator, StatusId.BATT_LEVEL, Int8ub)
        """Rough approximation of internal battery level in bars."""

        self.ext_batt_present: Status = Status(self.communicator, StatusId.EXT_BATT_PRESENT, Flag)  # type: ignore
        """Is an external battery connected?"""

        self.ext_batt_level: Status = Status(self.communicator, StatusId.EXT_BATT_LEVEL, Int8ub)
        """External battery power level in percent."""

        self.system_hot: Status = Status(self.communicator, StatusId.SYSTEM_HOT, Flag)  # type: ignore
        """Is the system currently overheating?"""

        self.system_busy: Status = Status(self.communicator, StatusId.SYSTEM_BUSY, Flag)  # type: ignore
        """Is the camera busy?"""

        self.quick_capture: Status = Status(self.communicator, StatusId.QUICK_CAPTURE, Flag)  # type: ignore
        """Is quick capture feature enabled?"""

        self.encoding_active: Status = Status(self.communicator, StatusId.ENCODING, Flag)  # type: ignore
        """Is the camera currently encoding (i.e. capturing photo / video)?"""

        self.lcd_lock_active: Status = Status(self.communicator, StatusId.LCD_LOCK_ACTIVE, Flag)  # type: ignore
        """Is the LCD lock currently active?"""

        self.video_progress: Status = Status(self.communicator, StatusId.VIDEO_PROGRESS, Int8ub)
        """When encoding video, this is the duration (seconds) of the video so far; 0 otherwise."""

        self.wireless_enabled: Status = Status(self.communicator, StatusId.WIRELESS_ENABLED, Flag)  # type: ignore
        """Are Wireless Connections enabled?"""

        self.pair_state: Status = Status(
            self.communicator, StatusId.PAIR_STATE, construct.Enum(Int8ub, params.PairState)
        )
        """What is the pair state?"""

        self.pair_type: Status = Status(
            self.communicator, StatusId.PAIR_TYPE, construct.Enum(Int8ub, params.PairType)
        )
        """The last type of pairing that the camera was engaged in."""

        self.pair_time: Status = Status(self.communicator, StatusId.PAIR_TIME, Int32ub)
        """	Time (milliseconds) since boot of last successful pairing complete action."""

        self.wap_scan_state: Status = Status(
            self.communicator, StatusId.WAP_SCAN_STATE, construct.Enum(Int8ub, params.WAPState)
        )
        """State of current scan for WiFi Access Points. Appears to only change for CAH-related scans."""

        self.wap_scan_time: Status = Status(self.communicator, StatusId.WAP_SCAN_TIME, Int8ub)
        """The time, in milliseconds since boot that the WiFi Access Point scan completed."""

        self.wap_prov_stat: Status = Status(
            self.communicator, StatusId.WAP_PROV_STAT, construct.Enum(Int8ub, params.WAPState)
        )
        """WiFi AP provisioning state."""

        self.remote_ctrl_ver: Status = Status(self.communicator, StatusId.REMOTE_CTRL_VER, Int8ub)
        """What is the remote control version?"""

        self.remote_ctrl_conn: Status = Status(self.communicator, StatusId.REMOTE_CTRL_CONN, Flag)  # type: ignore
        """Is the remote control connected?"""

        self.pair_state2: Status = Status(self.communicator, StatusId.PAIR_STATE2, GreedyBytes)  # type: ignore
        """Wireless Pairing State."""

        self.wlan_ssid: Status = Status(self.communicator, StatusId.WLAN_SSID, GreedyString(encoding="utf-8"))
        """Provisioned WIFI AP SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.ap_ssid: Status = Status(self.communicator, StatusId.AP_SSID, GreedyString(encoding="utf-8"))
        """Camera's WIFI SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.app_count: Status = Status(self.communicator, StatusId.APP_COUNT, Int8ub)
        """The number of wireless devices connected to the camera."""

        self.preview_enabled: Status = Status(self.communicator, StatusId.PREVIEW_ENABLED, Flag)  # type: ignore
        """Is preview stream enabled?"""

        self.sd_status: Status = Status(
            self.communicator, StatusId.SD_STATUS, construct.Enum(Int8ub, params.SDStatus)
        )
        """Primary Storage Status."""

        self.photos_rem: Status = Status(self.communicator, StatusId.PHOTOS_REM, Int32ub)
        """How many photos can be taken before sdcard is full?"""

        self.video_rem: Status = Status(self.communicator, StatusId.VIDEO_REM, Int32ub)
        """How many minutes of video can be captured with current settings before sdcard is full?"""

        self.num_group_photo: Status = Status(self.communicator, StatusId.NUM_GROUP_PHOTO, Int32ub)
        """How many group photos can be taken with current settings before sdcard is full?"""

        self.num_group_video: Status = Status(self.communicator, StatusId.NUM_GROUP_VIDEO, Int32ub)
        """Total number of group videos on sdcard."""

        self.num_total_photo: Status = Status(self.communicator, StatusId.NUM_TOTAL_PHOTO, Int32ub)
        """Total number of photos on sdcard."""

        self.num_total_video: Status = Status(self.communicator, StatusId.NUM_TOTAL_VIDEO, Int32ub)
        """Total number of videos on sdcard."""

        self.date_time: Status = Status(self.communicator, StatusId.DATE_TIME, GreedyString(encoding="utf-8"))
        """Current date/time (format: %YY%MM%DD%HH%MM%SS, all values in hex)."""

        self.ota_stat: Status = Status(
            self.communicator, StatusId.OTA_STAT, construct.Enum(Int8ub, params.OTAStatus)
        )
        """The current status of Over The Air (OTA) update."""

        self.download_cancel_pend: Status = Status(self.communicator, StatusId.DOWNLAD_CANCEL_PEND, Flag)  # type: ignore
        """Is download firmware update cancel request pending?"""

        self.mode_group: Status = Status(self.communicator, StatusId.MODE_GROUP, Int8ub)
        """Current mode group (deprecated in HERO8)."""

        self.locate_active: Status = Status(self.communicator, StatusId.LOCATE_ACTIVE, Flag)  # type: ignore
        """Is locate camera feature active?"""

        self.multi_count_down: Status = Status(self.communicator, StatusId.MULTI_COUNT_DOWN, Int32ub)
        """The current timelapse interval countdown value (e.g. 5...4...3...2...1...)."""

        self.space_rem: Status = Status(self.communicator, StatusId.SPACE_REM, Int64ub)
        """Remaining space on the sdcard in Kilobytes."""

        self.streaming_supp: Status = Status(self.communicator, StatusId.STREAMING_SUPP, Flag)  # type: ignore
        """Is streaming supports in current recording/flatmode/secondary-stream?"""

        self.wifi_bars: Status = Status(self.communicator, StatusId.WIFI_BARS, Int8ub)
        """WiFi signal strength in bars."""

        self.current_time_ms: Status = Status(self.communicator, StatusId.CURRENT_TIME_MS, Int32ub)
        """System time in milliseconds since system was booted."""

        self.num_hilights: Status = Status(self.communicator, StatusId.NUM_HILIGHTS, Int8ub)
        """The number of hilights in encoding video (set to 0 when encoding stops)."""

        self.last_hilight: Status = Status(self.communicator, StatusId.LAST_HILIGHT, Int32ub)
        """Time since boot (msec) of most recent hilight in encoding video (set to 0 when encoding stops)."""

        self.next_poll: Status = Status(self.communicator, StatusId.NEXT_POLL, Int32ub)
        """The min time between camera status updates (msec). Do not poll for status more often than this."""

        self.analytics_rdy: Status = Status(
            self.communicator, StatusId.ANALYTICS_RDY, construct.Enum(Int8ub, params.AnalyticsState)
        )
        """The current state of camera analytics."""

        self.analytics_size: Status = Status(self.communicator, StatusId.ANALYTICS_SIZE, Int32ub)
        """The size (units??) of the analytics file."""

        self.in_context_menu: Status = Status(self.communicator, StatusId.IN_CONTEXT_MENU, Flag)  # type: ignore
        """Is the camera currently in a contextual menu (e.g. Preferences)?"""

        self.timelapse_rem: Status = Status(self.communicator, StatusId.TIMELAPSE_REM, Int32ub)
        """How many min of Timelapse video can be captured with current settings before sdcard is full?"""

        self.exposure_type: Status = Status(
            self.communicator, StatusId.EXPOSURE_TYPE, construct.Enum(Int8ub, params.ExposureMode)
        )
        """Liveview Exposure Select Mode."""

        self.exposure_x: Status = Status(self.communicator, StatusId.EXPOSURE_X, Int8ub)
        """Liveview Exposure Select: y-coordinate (percent)."""

        self.exposure_y: Status = Status(self.communicator, StatusId.EXPOSURE_Y, Int8ub)
        """Liveview Exposure Select: y-coordinate (percent)."""

        self.gps_stat: Status = Status(self.communicator, StatusId.GPS_STAT, Flag)  # type: ignore
        """Does the camera currently have a GPS lock?"""

        self.ap_state: Status = Status(self.communicator, StatusId.AP_STATE, Flag)  # type: ignore
        """Is the WiFi radio enabled?"""

        self.int_batt_per: Status = Status(self.communicator, StatusId.INT_BATT_PER, Int8ub)
        """Internal battery level (percent)."""

        self.acc_mic_stat: Status = Status(
            self.communicator, StatusId.ACC_MIC_STAT, construct.Enum(Int8ub, params.ExposureMode)
        )
        """Microphone Accesstory status."""

        self.digital_zoom: Status = Status(self.communicator, StatusId.DIGITAL_ZOOM, Int8ub)
        """	Digital Zoom level (percent)."""

        self.wireless_band: Status = Status(
            self.communicator, StatusId.WIRELESS_BAND, construct.Enum(Int8ub, params.WiFiBand)
        )
        """Wireless Band."""

        self.dig_zoom_active: Status = Status(self.communicator, StatusId.DIG_ZOOM_ACTIVE, Flag)  # type: ignore
        """Is Digital Zoom feature available?"""

        self.mobile_video: Status = Status(self.communicator, StatusId.MOBILE_VIDEO, Flag)  # type: ignore
        """Are current video settings mobile friendly? (related to video compression and frame rate)."""

        self.first_time: Status = Status(self.communicator, StatusId.FIRST_TIME, Flag)  # type: ignore
        """Is the camera currently in First Time Use (FTU) UI flow?"""

        self.sec_sd_stat: Status = Status(
            self.communicator, StatusId.SEC_SD_STAT, construct.Enum(Int8ub, params.SDStatus)
        )
        """Secondary Storage Status (exclusive to Superbank)."""

        self.band_5ghz_avail: Status = Status(self.communicator, StatusId.BAND_5GHZ_AVAIL, Flag)  # type: ignore
        """Is 5GHz wireless band available?"""

        self.system_ready: Status = Status(self.communicator, StatusId.SYSTEM_READY, Flag)  # type: ignore
        """Is the system ready to accept commands?"""

        self.batt_ok_ota: Status = Status(self.communicator, StatusId.BATT_OK_OTA, Flag)  # type: ignore
        """Is the internal battery charged sufficiently to start Over The Air (OTA) update?"""

        self.video_low_temp: Status = Status(self.communicator, StatusId.VIDEO_LOW_TEMP, Flag)  # type: ignore
        """Is the camera getting too cold to continue recording?"""

        self.orientation: Status = Status(
            self.communicator, StatusId.ORIENTATION, construct.Enum(Int8ub, params.Orientation)
        )
        """The rotational orientation of the camera."""

        self.thermal_mit_mode: Status = Status(self.communicator, StatusId.THERMAL_MIT_MODE, Flag)  # type: ignore
        """Can camera use high resolution/fps (based on temperature)?"""

        self.zoom_encoding: Status = Status(self.communicator, StatusId.ZOOM_ENCODING, Flag)  # type: ignore
        """Is this camera capable of zooming while encoding (static value based on model, not settings)?"""

        self.flatmode_id: Status = Status(
            self.communicator, StatusId.FLATMODE_ID, construct.Enum(Int8ub, params.Flatmode)
        )
        """Current flatmode ID."""

        self.logs_ready: Status = Status(self.communicator, StatusId.LOGS_READY, Flag)  # type: ignore
        """	Are system logs ready to be downloaded?"""

        self.timewarp_1x_active: Status = Status(self.communicator, StatusId.TIMEWARP_1X_ACTIVE, Flag)  # type: ignore
        """Is Timewarp 1x active?"""

        self.video_presets: Status = Status(self.communicator, StatusId.VIDEO_PRESETS, Int32ub)
        """Current Video Preset (ID)."""

        self.photo_presets: Status = Status(self.communicator, StatusId.PHOTO_PRESETS, Int32ub)
        """Current Photo Preset (ID)."""

        self.timelapse_presets: Status = Status(self.communicator, StatusId.TIMELAPSE_PRESETS, Int32ub)
        """	Current Timelapse Preset (ID)."""

        self.presets_group: Status = Status(self.communicator, StatusId.PRESETS_GROUP, Int32ub)
        """Current Preset Group (ID)."""

        self.active_preset: Status = Status(self.communicator, StatusId.ACTIVE_PRESET, Int32ub)
        """Currently Preset (ID)."""

        self.preset_modified: Status = Status(self.communicator, StatusId.PRESET_MODIFIED, Int32ub)
        """Preset Modified Status, which contains an event ID and a preset (group) ID."""

        self.live_burst_rem: Status = Status(self.communicator, StatusId.LIVE_BURST_REM, Int32ub)
        """How many Live Bursts can be captured before sdcard is full?"""

        self.live_burst_total: Status = Status(self.communicator, StatusId.LIVE_BURST_TOTAL, Int32ub)
        """Total number of Live Bursts on sdcard."""

        self.capt_delay_active: Status = Status(self.communicator, StatusId.CAPT_DELAY_ACTIVE, Flag)  # type: ignore
        """Is Capture Delay currently active (i.e. counting down)?"""

        self.media_mod_mic_stat: Status = Status(
            self.communicator, StatusId.MEDIA_MOD_MIC_STAT, construct.Enum(Int8ub, params.MediaModMicStatus)
        )
        """Media mod State."""

        self.timewarp_speed_ramp: Status = Status(
            self.communicator, StatusId.TIMEWARP_SPEED_RAMP, construct.Enum(Int8ub, params.TimeWarpSpeed)
        )
        """Time Warp Speed."""

        self.linux_core_active: Status = Status(self.communicator, StatusId.LINUX_CORE_ACTIVE, Flag)  # type: ignore
        """Is the system's Linux core active?"""

        self.camera_lens_type: Status = Status(
            self.communicator, StatusId.CAMERA_LENS_TYPE, construct.Enum(Int8ub, params.MaxLensMode)
        )
        """Camera lens type (reflects changes to setting 162)."""

        self.video_hindsight: Status = Status(self.communicator, StatusId.VIDEO_HINDSIGHT, Flag)  # type: ignore
        """Is Video Hindsight Capture Active?"""

        self.scheduled_preset: Status = Status(self.communicator, StatusId.SCHEDULED_PRESET, Int32ub)
        """Scheduled Capture Preset ID."""

        self.scheduled_capture: Status = Status(self.communicator, StatusId.SCHEDULED_CAPTURE, Flag)  # type: ignore
        """Is Scheduled Capture set?"""

        self.creating_preset: Status = Status(self.communicator, StatusId.CREATING_PRESET, Flag)  # type: ignore
        """Is the camera in the process of creating a custom preset?"""

        self.media_mod_stat: Status = Status(
            self.communicator, StatusId.MEDIA_MOD_STAT, construct.Enum(Int8ub, params.MediaModStatus)
        )
        """Media Mode Status (bitmasked)."""

        self.turbo_mode: Status = Status(self.communicator, StatusId.TURBO_MODE, Flag)  # type: ignore
        """Is Turbo Transfer active?"""



# services.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Objects to nicely interact with BLE services, characteristics, and attributes.

TODO There needs to be significant future work to make this more useful. It would most likely
be possible to just use the bleak model of this directly but we need to account for other BLE adapters.
"""

import json
from dataclasses import dataclass, field, asdict
from typing import Dict, List, Union

from open_gopro.constants import UUID


def get_gopro_desc(uuid: str) -> Union[UUID, str]:
    """Attempt to retrieve a the name of a UUID from it's value.

    Args:
        uuid (str): string representation of UUID

    Returns:
        Union[UUID, str]: a UUID object if success, otherwise just the input string
    """
    try:
        return UUID(uuid.lower()).name
    except ValueError:
        return uuid


@dataclass
class Descriptor:
    """A charactersistic descriptor.

    Args:
        handle (int) : the handle of the attribute table that the descriptor resides at
        value (bytes) : the byte stream value of the descriptor
    """

    handle: int
    value: bytes

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(asdict(self), indent=4, default=str)


@dataclass
class Characteristic:
    """A BLE charactersistic.

    Args:
        handle (int) : the handle of the attribute table that the characteristic resides at
        uuid (UUID) : the UUID of the characteristic
        props (List[str]) : the characteristic's properties (READ, WRITE, NOTIFY, etc)
        name (str) : the characteristic's name
        value (bytes) : the byte stream value of the characteristic value
        descriptors (List[Descriptor], optional) : Any relevant descriptors if they exist
    """

    handle: int
    uuid: UUID
    props: List[str]
    name: str
    value: bytes
    descriptors: List[Descriptor] = field(default_factory=list)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return json.dumps(asdict(self), indent=4, default=str)


@dataclass
class Service:
    """A BLE service or grouping of Characteristics.

    Args:
        uuid (UUID) : the service's UUID
        name (str) : the service's name
        chars (Dict[str, Characteristic]) : the dictionary of characteristics, indexed by name
    """

    uuid: UUID
    name: str
    chars: Dict[UUID, Characteristic] = field(default_factory=dict)


class AttributeTable:
    """The attribute table to store / look up BLE services, characteristics, and attributes.

    Args:
        services (Dict[UUID, Service]): A dictionary of Services indexed by UUID..
    """

    def __init__(self, services: Dict[UUID, Service]) -> None:
        self.services = services

    def handle2uuid(self, handle: int) -> UUID:
        """Get a UUID from a handle.

        Args:
            handle (int): the handle to search for

        Raises:
            Exception: No characteristic was found at this handle

        Returns:
            UUID: The found UUID
        """
        for s in self.services.values():
            for c in s.chars.values():
                if c.handle == handle:
                    return c.uuid
        raise Exception(f"Matching UUID not found for handle {handle}")



# setup.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:51 UTC 2021

"""Package definition"""

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
VERSION = "0.5.7"
NAME = "open_gopro"
DESCRIPTION = "Open GoPro API and Examples"
URL = "https://gopro.github.io/OpenGoPro/python_sdk/"
EMAIL = "tcamise@gopro.com"
AUTHOR = "Tim Camise"

TEST_REQUIRED = ["pytest", "pytest-cov"]

REQUIRED = [
    "bleak==0.11.0",
    "construct>=2.10",
    "wrapt>=1.12.1",
    "requests",
    "rich>=9",
    "protobuf>=3",
    "betterproto",
]

here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()
# with io.open(os.path.join(here, "CHANGELOG.rst"), encoding="utf-8") as f:
#     long_description += "\n\n" + f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s: str):
        """Prints things in bold

        Args:
            s (str): Thing to print
        """
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=("tests", "examples", "docs")),
    entry_points={
        "console_scripts": [
            "gopro-demo=open_gopro.demos.demo:main",
            "gopro-photo=open_gopro.demos.photo:main",
            "gopro-video=open_gopro.demos.video:main",
            "gopro-stream=open_gopro.demos.stream:main",
        ]
    },
    install_requires=REQUIRED,
    test_suite="tests",
    tests_require=TEST_REQUIRED,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "Topic :: Communications",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    # setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



# wifi_commands.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Builds HTTP commands and defines parsing for anything sent over WiFi."""

import enum
import logging
from pathlib import Path
from abc import abstractmethod, ABC
from typing import Any, Optional, Dict, Callable

import wrapt
import requests

from open_gopro import params
from open_gopro.responses import GoProResp, ParserType, Parser
from open_gopro.constants import SettingId, StatusId

logger = logging.getLogger(__name__)


class WifiCommunicator(ABC):
    """Interface definition for a client to communicate via WiFi.

    This interface is used to build commands parse responses for:
    - :py:class:`open_gopro.wifi_commands.WifiCommands`
    - :py:class:`open_gopro.wifi_commands.WifiSettings`
    - :py:class:`open_gopro.wifi_commands.WifiStatuses`
    """

    @abstractmethod
    def get(self, url: str) -> GoProResp:
        """Send an HTTP GET request to a string endpoint.

        Args:
            url (str): endpoint not including GoPro base path

        Returns:
            GoProResp: GoPro response
        """
        raise NotImplementedError

    @abstractmethod
    def stream_to_file(self, url: str, file: Path) -> None:
        """Send an HTTP GET request to an Open GoPro endpoint to download a binary file.

        Args:
            url (str): endpoint URL
            file (Path): location where file should be downloaded to
        """
        raise NotImplementedError


# ======================================================Framework Builders==================================


def get_json(endpoint: str, response_parser: ParserType = None) -> Callable:
    """Build a command to send a GET HTTP request and return JSON response as GoProResp.

    Args:
        endpoint (str): endpoint to GET
        response_parser (ParserType, optional): Optional additional parsing to be done on the response data. Defaults to None.

    Returns:
        Callable: function to send GET request
    """

    @wrapt.decorator
    # pylint: disable = E, W
    def _wrapper(wrapped, instance, args, _):  # type: ignore
        logger.info(f'<----------- {wrapped.__name__} : {" ".join([str(x) for x in list(args)])}')

        if response_parser is not None:
            GoProResp.parser_map[endpoint] = response_parser

        # Build list of args as they should be represented in URL
        url_params = []
        for arg in args:
            if issubclass(type(arg), enum.Enum):
                url_params.append(arg.value)
            else:
                url_params.append(arg)
        url = endpoint.format(*url_params)
        # Send to camera
        response = instance.communicator.get(url)
        logger.info(f"-----------> \n{response}")
        return response

    return _wrapper


def get_binary(endpoint: str) -> Callable:
    """Build a command to send a GET HTTP request and receive a binary file.

    Args:
        endpoint (str): endpoint to GET

    Returns:
        Callable: function to send GET request
    """

    @wrapt.decorator
    # pylint: disable = E, W
    def _wrapper(wrapped, instance, _, kwargs):  # type: ignore
        logger.info(f'<----------- {wrapped.__name__} : {" ".join([str(x) for x in list(kwargs.values())])}')
        camera_file = kwargs["camera_file"]
        try:
            local_file = Path(kwargs["local_file"])
        except KeyError:
            local_file = Path(".") / f"{wrapped.__name__}-{camera_file}"

        url = endpoint.format(camera_file)
        # Send to camera
        try:
            instance.communicator.stream_to_file(url, local_file)
        except requests.exceptions.HTTPError as e:
            logger.error(repr(e))
        else:
            logger.info("-----------> SUCCESS")
        return local_file

    return _wrapper


@wrapt.decorator
# pylint: disable = E, W
def setter_cmd(wrapped, instance, args, _):  # type: ignore
    """Build a WiF i command that sets a Setting value."""
    logger.info(f"<----------- {wrapped.__name__} : {' '.join([str(a) for a in args])}")
    # Build url
    url = "gopro/camera/setting?setting_id={}&opt_value={}".format(instance.id.value, args[0].value)
    # Send to camera
    response = instance.communicator.get(url)
    if response is not None:
        logger.info(f"-----------> \n{response}")
    return response


# ======================================================Commands============================================


class ParseCameraState(Parser):
    """Additional parsing to do on received camera state."""

    def parse(self, json: Dict[str, Any]) -> Dict[Any, Any]:
        """Parse the raw state values into user friendly types / values.

        Args:
            json (Dict[str, Any]): input dict to parse

        Returns:
            Dict[Any, Any]: parsed output dict
        """
        parsed: Dict[Any, Any] = {}
        # Parse status and settings values into nice human readable things
        for (name, id_map) in [("status", StatusId), ("settings", SettingId)]:
            for k, v in json[name].items():
                id = id_map(int(k))
                try:
                    parser = GoProResp.parser_map[id]
                    # GreedyBytes, GreedyString, etc can't be built since they don't have a length
                    val = v if "Greedy" in str(parser) else parser.parse(parser.build(v))  # type: ignore
                except KeyError:
                    logger.warning(f"unparsed {name}: {id}")
                    val = v
                parsed[id] = val

        return parsed


class WifiCommands:
    """All of the Wifi commands.

    To be used as a delegate for a WifiCommunicator to build commands

    All of these return a GoProResp

    Args:
        communicator (WifiCommunicator): [description]
    """

    def __init__(self, communicator: WifiCommunicator):
        self.communicator = communicator

    @get_json("gp/gpControl/command/set_client_info")
    def set_third_party_client_info(self) -> GoProResp:
        """Flag as third party app."""

    @get_json("gopro/camera/digital_zoom?percent={}")
    def set_digital_zoom(self, zoom: int, /) -> GoProResp:
        """Set digital zoom in percent (0 to 100)."""

    @get_json("gopro/media/list")
    def get_media_list(self) -> GoProResp:
        """Get a list of media on the camera."""

    @get_json("gopro/version")
    def get_version(self) -> GoProResp:
        """Get Open GoPro version."""

    @get_json("gopro/camera/presets/get")
    def get_preset_status(self) -> GoProResp:
        """Get current Preset status."""

    @get_json("gopro/camera/presets/load?id={}")
    def set_preset(self, preset: params.Preset, /) -> GoProResp:
        """Set the active preset."""

    @get_json("gopro/camera/presets/set_group?id={}")
    def set_preset_group(self, group: params.PresetGroup, /) -> GoProResp:
        """Load a preset group."""

    @get_json("gopro/camera/stream/start")
    def start_preview_stream(self) -> GoProResp:
        """Start the preview stream."""

    @get_json("gopro/camera/stream/stop")
    def stop_preview_stream(self) -> GoProResp:
        """Stop the preview stream."""

    @get_json("gopro/media/info?path=100GOPRO/{}")
    def get_media_info(self, file: str, /) -> GoProResp:
        """Get media info for a file."""

    @get_json("gopro/camera/state", response_parser=ParseCameraState())
    def get_camera_state(self) -> GoProResp:
        """Get camera status and settings."""

    @get_json("gopro/media/turbo_transfer?p={}")
    def set_turbo_mode(self, enable: params.Toggle, /) -> GoProResp:
        """Enable / disable turbo mode."""

    @get_binary("gopro/media/gpmf?path=100GOPRO/{}")
    def get_gpmf_data(self, *, camera_file: str, local_file: Optional[Path] = None) -> Path:
        """Get GPMF data for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

    @get_binary("gopro/media/screennail?path=100GOPRO/{}")
    def get_screennail(self, *, camera_file: str, local_file: Optional[Path] = None) -> Path:
        """Get screennail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

    @get_binary("gopro/media/telemetry?path=100GOPRO/{}")
    def get_telemetry(self, *, camera_file: str, local_file: Optional[Path] = None) -> Path:
        """Get telemetry for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

    @get_binary("gopro/media/thumbnail?path=100GOPRO/{}")
    def get_thumbnail(self, *, camera_file: str, local_file: Optional[Path] = None) -> Path:
        """Get thumbnail for a file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """

    @get_binary("videos/DCIM/100GOPRO/{}")
    def download_file(self, *, camera_file: str, local_file: Optional[Path] = None) -> Path:
        """Download a video from the camera to a local file.

        If local_file is none, the output location will be the same name as the camera_file.

        Args:
            camera_file (str): filename on camera to operate on
            local_file (Optional[Path], optional): Location on computer to write output. Defaults to None.

        Returns:
            Path: Path to local_file that output was written to
        """


# =====================================================Settings============================================


class Setting(ABC):
    """An individual camera setting.

    Args:
        communicator (WifiCommunicator): Adapter to read write settings data
        id (SettingId): ID of setting
    """

    def __init__(self, communicator: WifiCommunicator, id: SettingId) -> None:
        self.id = id
        self.communicator = communicator

    @abstractmethod
    def set(self) -> GoProResp:
        """Set the value of the setting.

        This shall be implemented (and documented) by the subclass.

        Returns:
            GoProResp: Status of set
        """
        raise NotImplementedError


class WifiSettings:
    """The collection of all Settings.

    Args:
        communicator (WifiCommunicator): Adapter to read / write settings
    """

    def __init__(self, communicator: WifiCommunicator):
        self.communicator = communicator

        class Resolution(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, resolution: params.Resolution) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.resolution = Resolution(self.communicator, SettingId.RESOLUTION)
        """Resolution. Set with :py:class:`open_gopro.params.Resolution`"""

        class FOV(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, fov: params.FieldOfView) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.video_field_of_view = FOV(self.communicator, SettingId.VIDEO_FOV)
        """Video FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        self.photo_field_of_view = FOV(self.communicator, SettingId.PHOTO_FOV)
        """Photo FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        self.multi_shot_field_of_view = FOV(self.communicator, SettingId.MULTI_SHOT_FOV)
        """Multi-shot FOV. Set with :py:class:`open_gopro.params.FieldOfView`"""

        class FPS(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, fps: params.FPS) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.fps = FPS(self.communicator, SettingId.FPS)
        """Frames per second. Set with :py:class:`open_gopro.params.FPS`"""

        class MaxLensMode(Setting):  # pylint: disable=missing-class-docstring
            @setter_cmd
            def set(self, mode: params.MaxLensMode) -> GoProResp:  # pylint: disable=unused-argument
                ...

        self.max_lens_mode = MaxLensMode(self.communicator, SettingId.MAX_LENS_MOD)
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.params.MaxLensMode`"""



# response_generic_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: response_generic.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


class EnumResultGeneric(betterproto.Enum):
    RESULT_UNKNOWN = 0
    RESULT_SUCCESS = 1
    RESULT_ILL_FORMED = 2
    RESULT_NOT_SUPPORTED = 3
    RESULT_ARGUMENT_OUT_OF_BOUNDS = 4
    RESULT_ARGUMENT_INVALID = 5


@dataclass
class ResponseGeneric(betterproto.Message):
    result: "EnumResultGeneric" = betterproto.enum_field(1)



# util.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""Miscellaneous utilities for the GoPro package."""

import sys
import queue
import logging
import subprocess
from pathlib import Path
from typing import Dict, Type, Any, List, Optional, Union

logger = logging.getLogger(__name__)


def launch_vlc(location: Optional[Path]) -> None:
    """Launch VLC

    Args:
        location (Optional[Path]): path to VLC. If None, it will be automatically discovered
    """
    # This is a fairly lazy way to find VLC. We'll call it best effort.
    potential_vlc_locations: List[Union[Path, str]] = []
    command = "echo Invalid Platform"
    if "linux" in sys.platform.lower():
        potential_vlc_locations = [r'"/snap/bin/vlc"']
        command = 'su $(id -un 1000) -c "{} udp://@:8554 > /dev/null 2>&1 &"'
    elif "darwin" in sys.platform.lower():
        potential_vlc_locations = [r'"/Applications/VLC.app/Contents/MacOS/VLC"']
        command = "{} udp://@:8554 > /dev/null 2>&1 &"
    elif "win" in sys.platform.lower():
        potential_vlc_locations = [
            r'"/c/Program Files/VideoLAN/VLC/vlc.exe"',
            r'"/c/Program Files (x86)/VideoLAN/VLC/vlc.exe"',
            r'"C:\Program Files (x86)\VideoLAN\VLC\vlc.exe"',
            r'"C:\Program Files\VideoLAN\VLC\vlc.exe"',
        ]
        command = "{} udp://@:8554 &"

    potential_vlc_locations = potential_vlc_locations if location is None else [location]
    for vlc in potential_vlc_locations:
        response = cmd(command.format(vlc)).lower()

        if (
            " not " not in response
            and " no " not in response
            and " cannot " not in response
            and " unexpected " not in response
        ):
            logger.info("VLC launched")
            return

    logger.error("Failed to find VLC")


def scrub(obj: Any, bad_key: str) -> None:
    """Recursively scrub a dict or list to remove a given key in place.

    Args:
        obj (Any): dict or list to operate on. If neither, it will return immediately.
        bad_key (str): key to remove
    """
    if isinstance(obj, dict):
        for key in list(obj.keys()):
            if key == bad_key:
                del obj[key]
            else:
                scrub(obj[key], bad_key)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            if obj[i] == bad_key:
                del obj[i]
            else:
                scrub(obj[i], bad_key)
    else:
        # neither a dict nor a list, do nothing
        pass


def cmd(command: str) -> str:
    """Send a command to the shell and return the result.

    Args:
        command (str): command to send

    Returns:
        str: response returned from shell
    """
    logger.debug(f"Send cmd --> {command}")
    # Note: Ignoring unicode characters in SSIDs to prevent intermittent UnicodeDecodeErrors from occurring
    # while trying to connect to SSID when *any* AP is nearby that has unicode characters in the name
    response = (
        subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # type: ignore
        .stdout.read()
        .decode(errors="ignore")
    )
    logger.debug(f"Receive response --> {response}")

    return response


class Singleton:
    """To be subclassed to create a singleton class."""

    _instances: Dict[Type["Singleton"], Type["Singleton"]] = {}

    # pylint: disable=missing-return-doc, missing-return-type-doc
    def __new__(cls, *args, **kwargs):  # type: ignore
        """Check for existing instance."""
        if cls not in cls._instances:
            cls._instances[cls] = object.__new__(cls, *args, **kwargs)
        return cls._instances[cls]


class SnapshotQueue(queue.Queue):
    """A subclass of the default queue module to safely take a snapshot of the queue

    This is so we can access the elements (in a thread safe manner) without dequeuing them.
    """

    def snapshot(self) -> List[Any]:
        """Acquire the mutex, then return the queue's elements as a list.

        Returns:
            List[Any]: List of queue elements
        """
        with self.mutex:
            return list(self.queue)
